﻿using System.Data;
using Microsoft.Extensions.Logging;
using Microsoft.Graph;
using Microsoft.Graph.Me.SendMail;
using Microsoft.Graph.Models;
using Microsoft.Graph.Models.ODataErrors;
using Microsoft.Kiota.Abstractions.Authentication;

namespace Drive.Common.Email
{
    public class GraphEmailSender :
        ISendEmail
    {
        private readonly ILogger _logger;

        public GraphEmailSender(
            ILoggerFactory loggerFactory)
        {
            _logger = loggerFactory.CreateLogger("email-logger");
        }

        public async Task Send(
            EmailRequest request)
        {
            Message message = new()
            {
                From = new Recipient { EmailAddress = new EmailAddress { Address = request.From } },
                ToRecipients = request.To
                .Select(x => new Recipient { EmailAddress = new EmailAddress { Address = x.Trim() } })
                .ToList(),
                CcRecipients = request.Cc
                .Select(x => new Recipient { EmailAddress = new EmailAddress { Address = x.Trim() } })
                .ToList(),

                Subject = request.Subject,
                Body = new ItemBody
                {
                    ContentType = BodyType.Html,
                    Content = request.Content,
                }
            };

            foreach (EmailRequest.Attachment attachment in request.Attachments)
            {
                message.Attachments ??= new List<Attachment>();
                message.Attachments.Add(
                    new FileAttachment
                    {
                        Name = attachment.Name,
                        ContentId = attachment.ContentId,
                        ContentBytes = attachment.Content,
                        ContentType = attachment.ContentType,
                        IsInline = attachment.Inline,
                    });
            }

            GraphServiceClient client = new(
                new BaseBearerTokenAuthenticationProvider(
                    new DriveTokenProvider(request.AccessToken!)));

            try
            {
                await client.Me
                    .SendMail
                    .PostAsync(new SendMailPostRequestBody { Message = message });
            }
            catch (ODataError exception)
            {
                _logger.LogError($"ODataError: {exception.ResponseStatusCode} {exception.Error?.Code}: {exception.Error?.Message}", exception);
                throw;
            }
        }
    }

    public class DriveTokenProvider :
        IAccessTokenProvider
    {
        private readonly string _token;

        public DriveTokenProvider(
            string token)
        {
            _token = token;
            AllowedHostsValidator = new();
        }

        public AllowedHostsValidator AllowedHostsValidator { get; }

        public Task<string> GetAuthorizationTokenAsync(Uri uri, Dictionary<string, object>? additionalAuthenticationContext = null, CancellationToken cancellationToken = default)
        {
            return Task.FromResult(_token);
        }
    }
}
