﻿using System.Data;
using Dapper;

namespace Drive.Common.Security.Token
{
    public interface ITokenStorage
    {
        void Store(
            string userCode,
            string refreshToken,
            string accessToken);

        void Store(
            string userCode,
            string accessToken);

        string? Load(
            string userCode);

        void Revoke(
            string userCode);
    }

    public class TokenStorage
        : ITokenStorage
    {
        private readonly IClockProvider _clockProvider;
        private readonly IDbConnection _connection;

        public TokenStorage(
            IDbConnection connection,
            IClockProvider clockProvider)
        {
            _connection = connection;
            _clockProvider = clockProvider;
        }

        public void Store(
            string userCode,
            string refreshToken,
            string accessToken)
        {
            _connection.Execute(
                @"
MERGE UserRefreshToken AS target
USING (
        SELECT  OID,
                @refreshToken
        FROM    [User]
        WHERE   [User].Code = @userCode AND Active = 1
    ) AS source (UserId, RefreshToken)
ON (target.UserId = source.UserId)
    WHEN MATCHED THEN 
        UPDATE SET RefreshToken = source.RefreshToken, ModifiedAt = @modifiedAt
    WHEN NOT MATCHED THEN
        INSERT (UserId, RefreshToken, ModifiedAt) VALUES (source.UserId, source.RefreshToken, @modifiedAt);

MERGE UserAccessToken AS target
USING (
        SELECT  OID,
                @accessToken
        FROM    [User]
        WHERE   [User].Code = @userCode AND Active = 1
    ) AS source (UserId, AccessToken)
ON (target.UserId = source.UserId)
    WHEN MATCHED THEN 
        UPDATE SET AccessToken = source.AccessToken, ModifiedAt = @modifiedAt
    WHEN NOT MATCHED THEN
        INSERT (UserId, AccessToken, ModifiedAt) VALUES (source.UserId, source.AccessToken, @modifiedAt);",
                new { userCode, refreshToken, accessToken, modifiedAt = _clockProvider.UtcNow });
        }

        public void Store(
            string userCode,
            string accessToken)
        {
            _connection.Execute(
                @"
MERGE UserAccessToken AS target
USING (
        SELECT  OID,
                @accessToken
        FROM    [User]
        WHERE   [User].Code = @userCode AND Active = 1
    ) AS source (UserId, AccessToken)
ON (target.UserId = source.UserId)
    WHEN MATCHED THEN 
        UPDATE SET AccessToken = source.AccessToken, ModifiedAt = @modifiedAt
    WHEN NOT MATCHED THEN
        INSERT (UserId, AccessToken, ModifiedAt) VALUES (source.UserId, source.AccessToken, @modifiedAt);",
                new { userCode, accessToken, modifiedAt = _clockProvider.UtcNow });
        }

        public string? Load(
            string userCode)
        {
            return _connection.QuerySingleOrDefault<string>(
                @"
SELECT  RefreshToken
FROM    UserRefreshToken
        INNER JOIN [User] ON [User].OID = UserRefreshToken.UserId
WHERE   [User].Code = @userCode
        AND [User].Active = 1",
                new { userCode });
        }

        public void Revoke(
            string userCode)
        {
            _connection.Execute(
                @"
DELETE  UserRefreshToken
FROM    UserRefreshToken
        INNER JOIN [User] ON [User].OID = UserRefreshToken.UserId
WHERE   [User].Code = @userCode

DELETE  UserAccessToken
FROM    UserAccessToken
        INNER JOIN [User] ON [User].OID = UserAccessToken.UserId
WHERE   [User].Code = @userCode",
                new { userCode });
        }
    }
}
