﻿using System.Diagnostics;
using System.Reflection;

namespace Drive.Common
{
    [Serializable]
    [DebuggerDisplay("{Description} - {Value}")]
    public abstract class Enumeration<TEnumeration> : Enumeration<TEnumeration, int>
        where TEnumeration : Enumeration<TEnumeration>
    {
        protected Enumeration(int value, string description)
            : base(value, description)
        {
        }

        public static TEnumeration FromInt32(int value)
        {
            return FromValue(value);
        }

        public static bool TryFromInt32(int listItemValue, out TEnumeration? result)
        {
            return TryParse(listItemValue, out result);
        }
    }

    [Serializable]
    [DebuggerDisplay("{Description} - {Value}")]
    public abstract class Enumeration<TEnumeration, TValue> : IComparable<TEnumeration>, IEquatable<TEnumeration>
        where TEnumeration : Enumeration<TEnumeration, TValue>
        where TValue : IComparable
    {
        private readonly string _description;
        private readonly TValue _value;

        private static readonly Lazy<TEnumeration[]> _enumerations = new(GetEnumerations);

        protected Enumeration(TValue value, string description)
        {
            _value = value;
            _description = description;
        }

        public TValue Value
        {
            get { return _value; }
        }

        public string Description
        {
            get { return _description; }
        }

        public int CompareTo(TEnumeration? other)
        {
            TValue? value = other is not null ? other.Value : default;
            return Value.CompareTo(value);
        }

        public override sealed string ToString()
        {
            return Description;
        }

        public static TEnumeration[] GetAll()
        {
            return _enumerations.Value;
        }

        private static TEnumeration[] GetEnumerations()
        {
            Type enumerationType = typeof(TEnumeration);
            return enumerationType
                .GetFields(BindingFlags.Public | BindingFlags.Static | BindingFlags.DeclaredOnly)
                .Where(info => enumerationType.IsAssignableFrom(info.FieldType))
                .Select(info => info.GetValue(null))
                .Cast<TEnumeration>()
                .ToArray();
        }

        public override bool Equals(object? obj)
        {
            return Equals(obj as TEnumeration);
        }

        public bool Equals(TEnumeration? other)
        {
            return other is not null && Value.Equals(other.Value);
        }

        public override int GetHashCode()
        {
            return Value.GetHashCode();
        }

        public static bool operator ==(Enumeration<TEnumeration, TValue> left, Enumeration<TEnumeration, TValue> right)
        {
            return Equals(left, right);
        }

        public static bool operator !=(Enumeration<TEnumeration, TValue> left, Enumeration<TEnumeration, TValue> right)
        {
            return !Equals(left, right);
        }

        public static TEnumeration FromValue(TValue value)
        {
            return Parse(value, "value", item => item.Value.Equals(value));
        }

        public static TEnumeration Parse(string description)
        {
            return Parse(description, "description", item => item.Description == description);
        }

        private static bool TryParse(Func<TEnumeration, bool> predicate, out TEnumeration? result)
        {
            result = GetAll().FirstOrDefault(predicate);
            return result is not null;
        }

        private static TEnumeration Parse(object value, string description, Func<TEnumeration, bool> predicate)
        {

            if (!TryParse(predicate, out TEnumeration? result))
            {
                string message = string.Format("'{0}' is not a valid {1} in {2}", value, description, typeof(TEnumeration));
                throw new ArgumentException(message, nameof(value));
            }

            return result!;
        }

        public static bool TryParse(TValue value, out TEnumeration? result)
        {
            return TryParse(e => e.Value.Equals(value), out result);
        }

        public static bool TryParse(string description, out TEnumeration? result)
        {
            return TryParse(e => e.Description == description, out result);
        }
    }
}